
#line 6 "Io.mx"

/* Copyright (C) 1998-2001 Contributors 

* This program is free software; you can redistribute it and/or modify 
*  it under the terms of the GNU General Public License as published by  
*  the Free Software Foundation; either version 2 of the License, or  
*  (at your option) any later version.  
*
*  This program is distributed in the hope that it will be useful, 
*  but WITHOUT ANY WARRANTY; without even the implied warranty of   
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
*  GNU General Public License for more details.  
*
*  You should have received a copy of the GNU General Public License  
*  along with this program; if not, write to the Free Software  
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  
*/


#include	<stdio.h>
#include	<ctype.h>
#include 	<string.h>
#include	"Mx.h"
#include	"MxFcnDef.h"
#include <sys/types.h>
#include <sys/stat.h> 

File	files[M_FILES];
int	nfile = 0;

FILE *	ifile= 0;
FILE *	ofile= 0;
FILE *	ofile_body = 0;
FILE *	ofile_index = 0;

char *outputdir;
char *inputdir;
int outputdir_len;

int OutputDir( dir )
char *	dir;
{
    struct stat buf;
    if(stat(dir,&buf)==-1 || !(buf.st_mode&S_IFDIR)){
	fprintf(stderr,"Invalid target directory (%s), using \".\" instead\n",
		dir);
	return;
    }
    outputdir = StrDup(dir);
    outputdir_len = strlen( dir );
}
	
File *	GetFile(s,m)
char *	s;
int  m;
{
File *	f;
char *fname;
char *bname;

        bname = BaseName(s);
        fname = (char *)malloc(outputdir_len + strlen(bname)+strlen(dir2ext(m))+16);
	
	sprintf(fname, "%s/%s.%s", outputdir, BaseName(s), dir2ext(m));

	for( f= files; f < files + nfile; f++ ){
		if( strcmp(f->f_name, fname) == 0 ){
			return f;
		}
	}
	if( nfile == M_FILES )
		Fatal("GetFile", "Too many files");

	f->f_name= fname;

	f->f_str = fname+strlen(fname);;
	while(f->f_str>fname && f->f_str[-1]!='/') f->f_str--;

        f->f_tmp = (char *)malloc(strlen(outputdir)+strlen(bname)+strlen(dir2ext(m))+16); 
        sprintf(f->f_tmp,"%s/%s.%s ",outputdir,BaseName(s), dir2ext(m));
       
	f->f_mode= 0;

	nfile++;
	return f;
}

int	HasSuffix(name, suffix)
char *	name;
char *	suffix;
{
	if( strlen(name) <= strlen(suffix) )
		return 0;	
	return (strcmp(name + strlen(name) - strlen(suffix), suffix) == 0);
}

char	bname[128];

char *	BaseName(name)
char *	name;
{
    char *	b;
    char *	n;
    
    n=name;
    b= bname;
    /* @f /blah/blah indicates an absolute file name, 
       often used in Makefiles etc. ignore path
       @f blah/blah indicates a relative path,
       create sub-directories if necessary
    */
    if(name[0]=='/')
    	n= (strrchr(name, '/')? strrchr(name, '/') + 1: name);

    strcpy(bname,n);
    b=strrchr(bname,'.');
    if (b != NULL) *b= '\0';
    return bname;
}

FILE *fmustopen(fname,mode)
char *fname;
char *mode;
{
    char *p=fname; 
    char *tmp;
    if(*p=='/') p++;
    while((tmp=strchr(p,'/'))!=NULL){
	struct stat buf;	
        *tmp='\0';
        if(stat(fname,&buf)==-1 || !(buf.st_mode&S_IFDIR))
		if (mkdir(fname,S_IRWXU)==-1)
			Fatal("fmustopen:","Can't create directory %s\n",fname);
	p=tmp+1;
	*tmp='/';
    }
    return fopen(fname,mode);

}


char 	fname[128];

void	WriteFile(s, m)
char *	s;
int	m;
{
    File *	f;
	
    if( ofile )
	fclose(ofile);
    
    f= GetFile(s,m);
    if( (f->f_mode & m) == m ){
	ofile= fmustopen(f->f_tmp, "a");
	if(ofile==NULL)
	    Fatal("WriteFile","can't append to:%s",f->f_tmp);
    } else {
	f->f_mode |= m;
	ofile= fmustopen(f->f_tmp, "w");
	if(ofile==NULL)
	    Fatal("WriteFile","can't create:%s",f->f_tmp);
    }
}

void UpdateFiles()
{
    char *new, *old;
    char cmp[128];
    File *f;
    int status;
    
    if (ofile) fclose(ofile);

    for( f = files; f < files + nfile; f++) {
	if (ofile_index) {
            fclose(ofile_index); ofile_index=0;
            if (!itable_done) unlink(filename_index);
	}
	if (ofile_body) {
            fclose(ofile_body); ofile_body=0;
	    if (!itable_done) {
                FILE *fp = fopen(filename_body, "r+");
                /* TRICKY: remove the <base> tag in file header */
                char buf[1024], *start, *end;
                fread(buf, 1024, 1, fp);
                start = strstr(buf, "<BASE");
                end = strchr(start, '>');
                fseek(fp, start-buf, 0L);
                while(end-- >= start) fputc(' ', fp);
                fclose(fp);
 
                /* now use the body file as the output file */
                unlink(f->f_tmp);
                link(filename_body, f->f_tmp);
                unlink(filename_body);
	    }
	}
        sprintf(cmp,"cmp -s '%s' '%s'",f->f_tmp, f->f_name);
	status = system(cmp);
	if (status == 127) Fatal("UpdateFiles","can't execute %s",cmp);
	switch(status>>8){
	case 0: /* identical files, remove temporary file */
	    unlink(f->f_tmp);
	    break;
	case 1: /* different file */
    	    printf("File %s is modified \n",f->f_name);
	    if (rename(f->f_tmp, f->f_name)) perror("rename"); 
	    break;
	default: /* new file, move file */
    	    printf("Created file %s \n",f->f_name); 
	    if (rename(f->f_tmp, f->f_name)) perror("rename"); 
	    break;
	}
    }
}

void	ReadFile(name)
char *	name;
{
	char *p;

	if( !HasSuffix(name, ".mx") )
		Fatal("ReadFile", "Not a mx-file:%s", name);
	if( (ifile= fopen(name, "r")) == 0 ) 
		Fatal("ReadFile", "Can't process %s", name);
	p = (inputdir = strdup(name)) + strlen(name);
	while(--p >= inputdir && *p != '/');
	p[1] = 0;
	mx_file= name;
	mx_line= 1;
}

int	EofFile()
{
	return feof(ifile);
}

void	CloseFile()
{
	fclose(ifile);
	mx_file= 0;
	mx_line= 1;
}

#define MAXLINE 2048

static int  fullbuf=0;
static char linebuf[MAXLINE];

char *NextLine()
{
    mx_line++;
    if(fullbuf){ fullbuf = 0; return linebuf; }
    else {
	return fgets(linebuf,MAXLINE,ifile);
    }
}

char *PrevLine()
{
    mx_line--;
    fullbuf = 1; 
}


